# This file is part of librefdatool. librefdatool is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright (C) 2013 Javier D. Garcia-Lasheras
#

import numpy as np
from numpy import pi, log10
from scipy import signal
from matplotlib import pyplot as plt
from matplotlib import mlab

import sys, os, random
from PyQt4.QtCore import *
from PyQt4.QtGui import *

import matplotlib
from matplotlib.backends.backend_qt4agg import FigureCanvasQTAgg as FigureCanvas
from matplotlib.backends.backend_qt4agg import NavigationToolbar2QTAgg as NavigationToolbar
from matplotlib.figure import Figure

from scope import *


class Generator:
    '''This class instantiate a librefdatool signal generator object.
    This can be used for generating configurable signals in an easy way.
    The generated signals can be directly feeded to devices or scopes
    TBD: almost all!! Noise, DC, correct configuration...
    '''
    
    # Declare signals that will be sended to other classes
    stimulusUpdatedSignal = pyqtSignal()
    
    
    def __init__(self):
        
        self.x = np.ones(1)
        self.on_stimulus_draw()



    def generate_stimulus(self):
        '''This method returns a parametrized waveform signal:
        * waveform: assign signal geometry, with valid values:
        [doppler, chirp, sawtooth, square, gausspulse, sweep]
        * length: number of samples of the generated signal array
        * return: numeric array with the values of the generated signal
        TBD: the waveform must be parametrizable. Most templates doesnt work.
        Duration is in seconds
        Is Delay in samples or in seconds?
        TBD: we have to deal with discrete frequencies!!!!
        Forget the time, only samples and discrete frequencies should be involved!!!
        '''


        
        waveform = self.comboWaveform.currentText()
        length = int(self.textLength.text())
        # Parse the frequency value!!!
        # We should use as many array elements as frequencies given
        frequency = np.zeros(1)
        frequency[0] = float(self.textFrequency.text())   
        amplitude = float(self.textAmplitude.text())
        duration = float(self.textDuration.text())
        delay = int(self.textDelay.text())
        offset = float(self.textOffset.text())
        duty = float(self.textDuty.text())
        
        
        # Prepare the time array
        tInit=float(0)
        tFinal=duration
        t = np.linspace(tInit, tFinal, num=length)

        if  waveform == 'chirp':
            # Single chirp: Frequency-swept cosine generator.
            # (t, f0=0, t1=1, f1=100, method='linear', phi=0, qshape=None)
            # We set phi to 270 degrees because we want the signal to start in zero and then progressively rise
            wave = signal.chirp(t, f0=0, t1=tFinal, f1=frequency, method='linear', phi=270)

        elif waveform == 'sine':
            # sweep_poly is a frequency-swept cosine generator, with a time-dependent frequency.
            # We set a order zero polinomial, so we get just a cosine
            p = np.poly1d(frequency)
            wave = signal.sweep_poly(t, p, phi=270)

        elif waveform == 'sawtooth':
            # Sawtooth example: A 5 Hz waveform sampled at 500 Hz for 1 second:
            # t = np.linspace(0, 1, 500)
            # signal.sawtooth(2 * np.pi * 5 * t)
            wave = signal.sawtooth(2*np.pi*frequency*t, width=duty)

        elif waveform == 'square':
            # Square Wave
            wave = signal.square(2*np.pi*frequency*t, duty=duty)
            
        elif waveform == 'impulse':
            # Sweep Poly
            wave=np.zeros(length)
            wave[delay]=1

        elif waveform == 'step':
            # Sweep Poly
            wave=np.zeros(length)
            for ii in range(length-delay):
                wave[ii+delay]=1

        else :
            print('Not recognized waveform function: signal is zero')
            # Sweep Poly
            wave=np.zeros(length)


        for ii in range(len(wave)):
            wave[ii] = wave[ii]*amplitude + offset

        return wave


    def on_stimulus_draw(self):
        """ Redraws the figure
        """

        # clear the axes and redraw the plot anew
        #
       
        grid = self.cbGridStimulus.isChecked()
        self.x = self.generate_stimulus()
        
        print('Generate a basic signal using the generator')
        
        stimulusWave = Waveform(value = self.x,
                                label = 'Stimulus',
                                color = '#00ff00')
        
        scopeTime(self.figStimulus, [stimulusWave], grid)     
        
        self.canvasStimulus.draw()
        
        self.stimulusUpdatedSignal.emit()
        
    
    def create_stimulus_layout(self):
        # Create the mpl Figure and FigCanvas objects. 
        # 5x4 inches, 100 dots-per-inch
        #
        #self.dpi = 100
        #self.fig = Figure((5.0, 4.0), dpi=self.dpi)
        self.figStimulus = Figure()
        self.canvasStimulus = FigureCanvas(self.figStimulus)
        self.canvasStimulus.setParent(self.main_frame)
        
        # Since we have only one plot, we can use add_axes 
        # instead of add_subplot, but then the subplot
        # configuration tool in the navigation toolbar wouldn't
        # work.
        #
        #self.axes = self.fig.add_subplot(111)
        
        # Bind the 'pick' event for clicking on one of the bars
        #
        self.canvasStimulus.mpl_connect('pick_event', self.on_pick)
        
        # Create the navigation toolbar, tied to the canvas
        #
        self.mpl_toolbar = NavigationToolbar(self.canvasStimulus, self.main_frame)
        
        # Other GUI controls
        # 
        
        labelWaveform = QLabel('Waveform:')
        self.comboWaveform = QComboBox()
        self.comboWaveform.addItem("chirp")
        self.comboWaveform.addItem("sine")
        self.comboWaveform.addItem("sawtooth")
        self.comboWaveform.addItem("square")
        self.comboWaveform.addItem("impulse")
        self.comboWaveform.addItem("step")  
        self.comboWaveform.setEditable(True) 
        self.comboWaveform.lineEdit().setReadOnly(True)
        self.comboWaveform.lineEdit().setAlignment(Qt.AlignRight)
        for ii in range(self.comboWaveform.count()):
            self.comboWaveform.setItemData(ii, Qt.AlignRight, Qt.TextAlignmentRole)
        self.connect(self.comboWaveform, SIGNAL('currentIndexChanged(int)'), self.on_stimulus_draw)
             
        self.cbGridStimulus = QCheckBox("Show &Grid")
        self.cbGridStimulus.setChecked(True)
        self.connect(self.cbGridStimulus, SIGNAL('stateChanged(int)'), self.on_stimulus_draw)
        
        labelLength = QLabel('Sample Length:')
        self.textLength = QLineEdit()
        validatorLength = QIntValidator(1,100000)
        self.textLength.setValidator(validatorLength)
        self.textLength.setText('1000')
        self.textLength.setAlignment(Qt.AlignRight)
        self.connect(self.textLength, SIGNAL('editingFinished()'), self.on_stimulus_draw)
        
        labelFrequency = QLabel('Frequency:')
        self.textFrequency = QLineEdit()
        # QDoubleValidator.__init__ (self, float bottom, float top, int decimals)
        validatorFrequency = QDoubleValidator(0, 1000, 8)
        self.textFrequency.setValidator(validatorFrequency)
        self.textFrequency.setText('10')
        self.textFrequency.setAlignment(Qt.AlignRight)
        self.connect(self.textFrequency, SIGNAL('editingFinished()'), self.on_stimulus_draw)        

        labelAmplitude = QLabel('Amplitude:')
        self.textAmplitude = QLineEdit()
        # QDoubleValidator.__init__ (self, float bottom, float top, int decimals)
        validatorAmplitude = QDoubleValidator(0, 1000, 8)
        self.textAmplitude.setValidator(validatorAmplitude)
        self.textAmplitude.setText('0.5')
        self.textAmplitude.setAlignment(Qt.AlignRight)
        self.connect(self.textAmplitude, SIGNAL('editingFinished()'), self.on_stimulus_draw)	  
        
        labelDuration = QLabel('Duration:')
        self.textDuration = QLineEdit()
        # QDoubleValidator.__init__ (self, float bottom, float top, int decimals)
        validatorDuration = QDoubleValidator(0, 1000, 8)
        self.textDuration.setValidator(validatorDuration)
        self.textDuration.setText('1')
        self.textDuration.setAlignment(Qt.AlignRight)
        self.connect(self.textDuration, SIGNAL('editingFinished()'), self.on_stimulus_draw)       
        
        labelDelay = QLabel('Delay:')
        self.textDelay = QLineEdit()
        validatorDelay = QIntValidator(0, 1000)
        self.textDelay.setValidator(validatorDelay)
        self.textDelay.setText('0')
        self.textDelay.setAlignment(Qt.AlignRight)
        self.connect(self.textDuration, SIGNAL('editingFinished()'), self.on_stimulus_draw)

        labelOffset = QLabel('Offset:')
        self.textOffset = QLineEdit()
        # QDoubleValidator.__init__ (self, float bottom, float top, int decimals)
        validatorOffset = QDoubleValidator(-1000, 1000, 8)
        self.textOffset.setValidator(validatorOffset)
        self.textOffset.setText('0')
        self.textOffset.setAlignment(Qt.AlignRight)
        self.connect(self.textOffset, SIGNAL('editingFinished()'), self.on_stimulus_draw)

        labelDuty = QLabel('Duty:')
        self.textDuty = QLineEdit()
        # QDoubleValidator.__init__ (self, float bottom, float top, int decimals)
        validatorDuty = QDoubleValidator(0, 1, 8)
        self.textDuty.setValidator(validatorDuty)
        self.textDuty.setText('0')
        self.textDuty.setAlignment(Qt.AlignRight)
        self.connect(self.textDuty, SIGNAL('editingFinished()'), self.on_stimulus_draw)
     
        
        #
        # Layout with box sizers
        # 
        hboxWaveform = QHBoxLayout()
        hboxWaveform.addWidget(labelWaveform)
        hboxWaveform.addWidget(self.comboWaveform)
        hboxWaveform.setStretch(0,1)
        hboxWaveform.setStretch(1,1)
        
        hboxLength = QHBoxLayout()
        hboxLength.addWidget(labelLength)
        hboxLength.addWidget(self.textLength)
        hboxLength.setStretch(0,1)
        hboxLength.setStretch(1,1)
        
        hboxFrequency = QHBoxLayout()
        hboxFrequency.addWidget(labelFrequency)
        hboxFrequency.addWidget(self.textFrequency)
        hboxFrequency.setStretch(0,1)
        hboxFrequency.setStretch(1,1)
        
        hboxAmplitude = QHBoxLayout()
        hboxAmplitude.addWidget(labelAmplitude)
        hboxAmplitude.addWidget(self.textAmplitude)
        hboxAmplitude.setStretch(0,1)
        hboxAmplitude.setStretch(1,1)
        
        hboxDuration = QHBoxLayout()
        hboxDuration.addWidget(labelDuration)
        hboxDuration.addWidget(self.textDuration)
        hboxDuration.setStretch(0,1)
        hboxDuration.setStretch(1,1)
        
        hboxDelay = QHBoxLayout()
        hboxDelay.addWidget(labelDelay)
        hboxDelay.addWidget(self.textDelay)
        hboxDelay.setStretch(0,1)
        hboxDelay.setStretch(1,1)
        
        hboxOffset = QHBoxLayout()
        hboxOffset.addWidget(labelOffset)
        hboxOffset.addWidget(self.textOffset)
        hboxOffset.setStretch(0,1)
        hboxOffset.setStretch(1,1)
        
        hboxDuty = QHBoxLayout()
        hboxDuty.addWidget(labelDuty)
        hboxDuty.addWidget(self.textDuty)
        hboxDuty.setStretch(0,1)
        hboxDuty.setStretch(1,1)
        
        
        vboxLeft = QVBoxLayout()
        vboxLeft.addLayout(hboxWaveform)
        vboxLeft.addLayout(hboxLength)
        vboxLeft.addLayout(hboxFrequency)
        vboxLeft.addLayout(hboxAmplitude)
        
        groupBoxLeft = QGroupBox()
        groupBoxLeft.setLayout(vboxLeft)
        
        vboxRight = QVBoxLayout()
        vboxRight.addLayout(hboxDuration)
        vboxRight.addLayout(hboxDelay)
        vboxRight.addLayout(hboxOffset)
        vboxRight.addLayout(hboxDuty)
        
        groupBoxRight = QGroupBox()
        groupBoxRight.setLayout(vboxRight)
        
        hboxStimulusControl = QHBoxLayout()
        hboxStimulusControl.addWidget(groupBoxLeft)
        hboxStimulusControl.addWidget(groupBoxRight)
        
        groupBoxStimulusControl = QGroupBox('Stimulus Control')
        groupBoxStimulusControl.setLayout(hboxStimulusControl)
        groupBoxStimulusControl.isCheckable()
        
        
        vboxMain = QVBoxLayout()
        vboxMain.addWidget(self.canvasStimulus)
        vboxMain.addWidget(self.mpl_toolbar)
        vboxMain.addWidget(self.cbGridStimulus)
        vboxMain.addWidget(groupBoxStimulusControl)
        
        self.on_stimulus_draw()
               
        return vboxMain

