# This file is part of librefdatool. librefdatool is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright (C) 2013 Javier D. Garcia-Lasheras
#

import sys
import os
import subprocess
import csv

import numpy as np
from scipy import signal

from bitstring import BitArray

from Verilog_VCD import *

from snippets import *


class Simcore:
    '''This is a temporary class
    Should it be removed as the development advances??
    '''

    def simfilter(self, b, a, x, 
            structure, language,
            engine, name,
            busX, busY, busC, 
            scalingX, scalingC):
        
        self.name = name
        self.structure = structure
        self.language = language
        self.engine = engine
        
        #TODO: debugging prints
        print('structure', structure)
        print('language', language)
        print('engine', engine)
        print('name', name)

        self.busX = busX
        self.busY = busY
        self.busC = busC

        self.scalingX = scalingX
        self.scalingC = scalingC
        
        # Define order values
        self.M = len(b)
        self.N = len(a)


        # Convert coef to fixed point

        self.coefB = b
        self.icoefB = self._quantizer(b, self.busC)
        
        # The A coefficients calculated with scipy uses a different convention and sign changes!!!!    
        a = -1 * a
        
        self.coefA = a
        self.icoefA = self._quantizer(a, self.busC)
        
        print('Coef B')
        print(self.icoefB)
        print('Coef A')
        print(self.icoefA)
        

        # Set DUT
        self._generateDUT()
        self._generateTB(x)

        return self._simulation(x)


    #***********************************************#
    #    Generate DUT
    #***********************************************#


    def _generateDUT(self):

            # All the products has an extra integer (the 2^1 factor)
            # A fixed point number is comprised by integer (P) and fractional bits (Q)
            # When you calculate the C product of A and B, then => PC = PA + PB; QC = QA + QB;
            # We are dealing with P=1 for all our numbers, so we can discard the resulting MSB.
            # By this way, the product of A and B limited to [-1,1) has a width of WC = WA + WB - 1
            
            # NOTE: we need to consider the possibility of P>1 !! (there are bigger than 1 coefficients) 
            # We are going to allow the full fixed point specification for X, Y and C!!
            
            # In the Sum S = A + B, the worst case for P is 1 + max(PA, PB).
            # For QS, we have the must of QA being equal to QB, so QA = QB = QS
            # As we have PA = PB = 1, PS would be 1 if there is not overflow. 
            # In any case, we can discard the overflow and assume that WC = WA = WS
            
            # In order to detect the overflow in the sum S = A + B, WA = WB:
            # If MSB(A) != MSB(B) => overflow cannot exist
            # if MSB(A)=MSB(B)=1 && MSB(S)=0 => overflow in the minimum range
            # if MSB(A)=MSB(B)=0 && MSB(S)=1 => overflow in the maximum range

        busX = self.busX[0] + self.busX[1]
        busY = self.busY[0] + self.busY[1]
        busC = self.busC[0] + self.busC[1]
        
        M = self.M
        N = self.N
        
        hdl = snippets()
        
        if self.language == 'VHDL':  

            structure = self.structure
            name = self.name

            # Code generation for a Transposed Direct Form FIR
            hdlFile = open('{!s}.vhd'.format(name),'w')

            hdlFile.write(hdl.libraries())
            hdlFile.write('\n')
            
            hdlFile.write(hdl.entity(name, busX, busY))
            hdlFile.write('\n')
            
            hdlFile.write(hdl.architectureHeader(name, structure))
            hdlFile.write('\n')
            
            
            # *** DECLARE INTERNAL SIGNALS ***
            
            hdlFile.write(hdl.signalZ(self.M, self.N, busX, busY, busC, structure))
            hdlFile.write('\n')  
            
            hdlFile.write(hdl.signalC(self.M, self.N, busC, structure))
            hdlFile.write('\n')     

            hdlFile.write(hdl.signalAdders(self.M, self.N, busX, busY, busC, structure))
            hdlFile.write('\n')      
            
            hdlFile.write(hdl.signalProducts(self.M, self.N, busX, busY, busC, structure))
            hdlFile.write('\n')    
            
            hdlFile.write(hdl.signalFeedback(self.M, self.N, busX, busY, busC, structure))
            hdlFile.write('\n')   
       
        
            hdlFile.write("-- Begin Architecture\n")
            hdlFile.write("begin\n\n")

            hdlFile.write(hdl.sequentialBlock(M, N, clkEdge='posedge', rstActive='high'))
            hdlFile.write("\n")

            hdlFile.write('-- Arithmetics\n')
            hdlFile.write("\n")

            hdlFile.write(hdl.assignCoeff(M, N, self.icoefB, self.coefB,
                                           self.icoefA, self.coefA, structure))
            hdlFile.write("\n")

            # Products
            hdlFile.write(hdl.arithProduct(M, N,
                                           busX, busY, busC, 
                                           structure))
            hdlFile.write("\n")            

            # Sums
            hdlFile.write(hdl.arithAdders(M, N, structure))
            hdlFile.write("\n") 


            hdlFile.write(hdl.connectBlocks(M, N, 
                                            self.busX, self.busY, self.busC,
                                            structure))
            hdlFile.write('\n') 

            hdlFile.write('end {!s};\n'.format(structure))
            hdlFile.close()
            


        elif self.language == 'Verilog':  

            structure = self.structure

            # Verilog Code generation for a Transposed Direct Form FIR
            hdlFile = open('{!s}.v'.format(structure),'w')

            hdlFile.write("// This is a transposed direct form test\n\n")
            
            hdlFile.write("`timescale 1ns/10ps\n\n")
            
            hdlFile.write("module {!s} (\n".format(structure))
            hdlFile.write("    clk,\n")
            hdlFile.write("    rst,\n")
            hdlFile.write("    sig_in,\n")
            hdlFile.write("    sig_out\n")
            hdlFile.write(");\n\n")

            
            hdlFile.write('// Architecture depends on selected structure\n')
            hdlFile.write('// It would be nice to have target optimizations too in the future\n')
            hdlFile.write('//\n')
            hdlFile.write('// example, transposed direct form for FIR:\n')
            hdlFile.write('//\n')
            hdlFile.write('//              z^-1  z^-1             z^-1  z^-1\n')
            hdlFile.write('//            o-->--o-->--o-->- - - --o-->--o-->--o-->--o y[n]\n')
            hdlFile.write('//            |     |     |           |     |     | \n')
            hdlFile.write('//            ^Bm   ^Bm-1 ^Bm-2       ^B2   ^B1   ^B0\n')
            hdlFile.write('//            |     |     |           |     |     |\n')
            hdlFile.write('// x[n] o-->--o-->--o-->--o-->- - - --o-->--o-->--\n')
            hdlFile.write('//\n\n')
            
            hdlFile.write("input clk;\n")
            hdlFile.write("input rst;\n")
            hdlFile.write("input signed [{!s}:0] sig_in;\n".format(self.busX-1))
            hdlFile.write("output signed [{!s}:0] sig_out;\n".format(self.busY-1))
            
            hdlFile.write("// Z^-1 delay blocks\n")
            hdlFile.write('reg signed [{!s}:0] z [0:{!s}];\n'.format(self.busX + self.busC -1, len(self.icoef)-1))
            hdlFile.write('wire signed [{!s}:0] z_next [0:{!s}];\n'.format(self.busX + self.busC -1, len(self.icoef)-1))
            hdlFile.write("\n")
            
            hdlFile.write("// Filter constants\n")
            for ii in range(self.M):               
                hdlFile.write('wire signed [{!s}:0] b{!s};\n'.format(self.busC -1, ii))
            hdlFile.write("\n")

            hdlFile.write("// Filter Adders\n")
            for ii in range(self.M-1):
                hdlFile.write('wire signed [{!s}:0] s{!s};\n'.format(self.busX + self.busC -1, ii))
            hdlFile.write("\n")

            hdlFile.write("// Filter Products\n")
            for ii in range(self.M):
                hdlFile.write('wire signed [{!s}:0] p{!s};\n'.format(self.busX + self.busC -1, ii))
            hdlFile.write("\n")  
            
            
            hdlFile.write("always @(posedge clk) begin: {!s}\n".format(structure))
            hdlFile.write("    if (rst == 1'b1) begin\n")
            for ii in range(self.M-1):
                hdlFile.write('        z[{!s}] <= 0;\n'.format(ii))
            hdlFile.write("    end\n")
            hdlFile.write("    else begin\n")
            for ii in range(self.M-1):
                hdlFile.write('        z[{!s}] <= z_next[{!s}];\n'.format(ii, ii))
            hdlFile.write("    end\n")
            hdlFile.write("end\n")
            hdlFile.write("\n")
            
            hdlFile.write('// Arithmetics\n')
            hdlFile.write("\n")
            for ii in range(self.M):
                hdlFile.write('assign b{!s} = {!s}\'b{!s}; // {!s}\n'.format(ii, self.busC, self.icoef[ii], self.coef[ii]))
            hdlFile.write("\n")  
            for ii in range(self.M):
                hdlFile.write('assign p{!s} = sig_in * b{!s};\n'.format(ii, ii))
            hdlFile.write("\n")  
            for ii in range(self.M-1):
                hdlFile.write('assign s{!s} = p{!s} + z[{!s}];\n'.format(ii, ii, ii))
            hdlFile.write("\n\n")  

            hdlFile.write('// Signal link, trunk, overflow...\n')
            for ii in range(self.M):
                if ii == 0:
                    # Why always is a Zero as MSB in si? --> this is because the input is always a positive number?????
                    hdlFile.write('assign sig_out = s{!s}[{!s}:{!s}];\n'.format(ii, self.busX + self.busC -1, self.busX + self.busC - self.busY))
                elif ii == (len(self.coef)-1):
                    hdlFile.write('assign z_next[{!s}] = p{!s};\n'.format(ii-1, ii))
                else:
                    hdlFile.write('assign z_next[{!s}] = s{!s};\n'.format(ii-1, ii))
            hdlFile.write("\n\n")  
            
            hdlFile.write("endmodule\n")
            hdlFile.write("\n")
           
            hdlFile.close()

 
        else:
            print("Error: simulation language not recognized\n");

    
        return 0


    def _generateTB(self, x):
        
        #if self.language == 'vhdl': 
            # TBD: the process of input signal and output value parse
            # must be aligned and configurable
            # This Verilog file is common for verilog and vhdl DUTs in IVerilog
            # If we want to add support for more simulators, then would be necessary VHDL

            # Code generation for the testbench - myhdl link 
        
        hdl = snippets()

        busX = self.busX[0] + self.busX[1]
        busY = self.busY[0] + self.busY[1]
        
        name = self.name
        
        xb = self._quantizer(x, self.busX, staircase='midtread')
     
        if self.engine == 'Icarus Verilog':
            
            tbFile = open('tb_{!s}.v'.format(name),'w')
            tbFile.write(hdl.verilogTestBench(name, xb, x, busX, busY))
            tbFile.close()
            
        
        elif self.engine == 'GHDL':    
            tbFile = open('tb_{!s}.vhd'.format(name),'w')
            tbFile.write(hdl.vhdlTestBench(name, xb, x, busX, busY))
            tbFile.close()
            



    def _quantizer(self, signal, bus, xm=1, staircase= 'midtread'):
        
        # Normalization to the [-1, 1) input range 
        signalNorm = signal/xm 
        
        # Generate the fixed point code in two's complement
        # 
        # xb = -a0*2^0 + a1*2^(-1) + a2*2^(-2) + ... + aB*2^(-B)
        # bus = B + 1; 
        
        B = bus[1] + bus[0] - 1
        
        if staircase == 'midtread':
            delta = 0.5
        elif staircase == 'midriser':
            delta = 0
        else:
            print("Staircase model for quantizer unknown: using default")
            delta = 0.5
        
        # Convert the normalized signal to a integer value
        signalScaled = signalNorm*(2**(bus[1])) + delta
        signalInt = np.floor(signalScaled)
        
        
        # Trunk if the integers cannot be represented using the bus width
        # Then assign the correspondent binary code
        xBinary = []
        for ii in range(len(signal)):
            if signalInt[ii] > (2**B -1):
                signalInt[ii] = 2**B - 1
            elif signalInt[ii] < (-1 * 2**B):
                signalInt[ii] = -1 * 2**B
            xBinary.append(np.binary_repr(int(signalInt[ii]), width=B+1))
        
        return xBinary
        
   
    def _simulation(self, stimulus):

        name = self.name
        totalSamples = len(stimulus)
        timeStep = 50
        
        
        if self.language == 'Verilog':
            dutExtension = 'v'
        elif self.language == 'VHDL':
            dutExtension = 'vhd'
        else:
            print('Unknown model: assuming verilog')
            dutExtension = 'v'        
        
        # Check if verilog/vhdl filter description is available TBD: make it a variable!!!!!!!!!!!!!!!!
        if not os.path.isfile('{!s}.{!s}'.format(name, dutExtension)):
            print('File not found: Run generation first')
            return None
        # Clean design
        if os.path.isfile('{!s}'.format(name)):
            print('Existing compiled design: erase!! ')
            cmd = 'rm {!s}'.format(name)
            os.system(cmd)
        
        
        if self.engine == 'Icarus Verilog': 
            # Running Icarus verilog
            # ****** TBD: Set the dut file extension according with the HDL language  
            cmd = 'iverilog -g2012 -o {!s} {!s}.{!s} tb_{!s}.v'.format(name, name, dutExtension, name)
            os.system(cmd)
    
            cmd = 'vvp {!s}'.format(name)
    
            proc = subprocess.Popen(cmd, stdout=subprocess.PIPE, shell=True)
            (out, err) = proc.communicate()
            
            
        elif self.engine == 'GHDL':
            #Running GHDL
            stopTime = (5 + totalSamples)*timeStep
            
            # Analyze the UUT
            cmd = 'ghdl -a {!s}.vhd'.format(name)
            proc = subprocess.Popen(cmd, stdout=subprocess.PIPE, shell=True)
            (out, err) = proc.communicate()    

            # Analyze the TestBench
            cmd = 'ghdl -a tb_{!s}.vhd'.format(name)
            proc = subprocess.Popen(cmd, stdout=subprocess.PIPE, shell=True)
            (out, err) = proc.communicate()  
          
            # Build an executable file    
            cmd = 'ghdl -e tb_{!s}'.format(name)
            proc = subprocess.Popen(cmd, stdout=subprocess.PIPE, shell=True)
            (out, err) = proc.communicate()
            print('GHDL executed')    
            
            # Run the executable    
            cmd = 'ghdl -r tb_{!s} --vcd={!s}.vcd --stop-time={!s}ns'.format(name, name, stopTime)
            proc = subprocess.Popen(cmd, stdout=subprocess.PIPE, shell=True)
            (out, err) = proc.communicate()
            print('GHDL runned')     

        
        # *****************************************************************************
        #vcdDebug = parse_vcd('{!s}.vcd'.format(name), 
        #                only_sigs=1
        #                )  
        
        #print(vcdDebug)
        
        
        if self.engine == 'Icarus Verilog':
            topName = 'tb_{!s}'.format(name)
            timeScale = 1
        elif self.engine == 'GHDL':
            topName = 'uut'
            timeScale = 1000000
        
        print('Parsing VCD')   
        vcd = parse_vcd('{!s}.vcd'.format(name), 
                        #use_stdout=1,
                        siglist=['{!s}sig_out[{!s}:0]'.format(
                                 topName, self.busY[0] + self.busY[1] - 1)]
                        )  
        for key in sorted(vcd):
            vcdKey = key
        
        print('Parsed VCD')
        print(vcd[vcdKey]['tv'])  
        
        
        vcdParsedOutput = []
        #vcdBinaryOutput = []
         
        outCurrentTime = 0
        vcdCurrentValue = vcd[vcdKey]['tv'][0][1]
        binOffset = self.busY[0] + self.busY[1] - len(vcdCurrentValue)
        for jj in range(binOffset):
            vcdCurrentValue = '0' + vcdCurrentValue
        
        for ii in range(len(vcd[vcdKey]['tv'])):
            vcdNextTime = vcd[vcdKey]['tv'][ii][0]
            vcdNextTime = vcdNextTime/timeScale
            vcdNextValue = vcd[vcdKey]['tv'][ii][1]
            while vcdNextTime > outCurrentTime:
                a = BitArray(bin=vcdCurrentValue)
                #vcdBinaryOutput.append(vcdCurrentValue)
                vcdParsedOutput.append(a.int/float(2**(self.busY[1])))
                vcdCurrentValue = vcdNextValue
                binOffset = self.busY[0] + self.busY[1] - len(vcdCurrentValue)
                for jj in range(binOffset):
                    vcdCurrentValue = '0' + vcdCurrentValue
                outCurrentTime += timeStep
        
        print('Signal reconstructed')   
        
        # remove the output samples from filter initialization        
        for ii in range(5):
            vcdParsedOutput.pop(0)
            #vcdBinaryOutput.pop(0)    
        
        connect_sample_out = np.zeros(len(vcdParsedOutput))
        for ii in range(len(vcdParsedOutput)):
            connect_sample_out[ii] = vcdParsedOutput[ii]
            
        return connect_sample_out
    
    

