# This file is part of librefdatool. librefdatool is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright (C) 2013 Javier D. Garcia-Lasheras
#

import warnings

from simcore import *

import numpy as np
from numpy import pi, log10
from scipy import signal
from matplotlib import pyplot as plt


import sys, os, random
from PyQt4.QtCore import *
from PyQt4.QtGui import *

import matplotlib
from matplotlib.backends.backend_qt4agg import FigureCanvasQTAgg as FigureCanvas
from matplotlib.backends.backend_qt4agg import NavigationToolbar2QTAgg as NavigationToolbar
from matplotlib.figure import Figure

from scope import *



class Simulator(Simcore):
    '''This class is the graphical interface to HDL simulator
    '''
    
    # Declare signals that will be sended to other classes
    simulatorUpdatedSignal = pyqtSignal()
    simulatorBeginSignal = pyqtSignal()
    simulatorEndSignal = pyqtSignal()
    
        

    def on_simulator_draw(self):
        """ Redraws the figure
        """
        
        # create waveforms
        HDLWave = Waveform(value = self.yhdl,
                           label = 'HDL',
                           color = '#ff0000')
        
        floatWave = Waveform(value = self.yfloat,
                             label = 'float',
                             color = '#0000ff')

        simulatorWaves = [floatWave, HDLWave]
        
        
        selectedPlot = str(self.comboSimulatorScope.currentText())
        
        grid = self.cbGridSimulator.isChecked()
        
        if selectedPlot == 'Time Plot':
            scopeTime(self.figSimulator, simulatorWaves, grid)
        elif selectedPlot == 'Error Plot':
            scopeError(self.figSimulator, self.yfloat, self.yhdl, grid)
        elif selectedPlot == 'Power Spectrum':
            scopePower(self.figSimulator, simulatorWaves, grid)
        else:
            scopeTime(self.figSimulator, simulatorWaves, grid)
            
        # Change pushbutton state
        self.pbUpdateFilter.setDisabled(True)
        self.pbUpdateFilter.setStyleSheet('QPushButton {background-color: #00FF00; color: #FFFFFF}')
   
        self.canvasSimulator.draw()
            

    def execute_simulation(self, b, a, x,
                           name, language,
                           engine, structure,
                           busX, busY, busC,
                           scalingX, scalingC):
        """ Redraws the figure
        """


        # Get the coefficients
        warningMessage = ''
        warnings.simplefilter('error')
        try:
            self.simulatorBeginSignal.emit()
            self.yfloat = signal.lfilter(b, a, x)
            self.yhdl = self.simfilter(b, a, x,
                                       structure, language,
                                       engine, name,
                                       busX, busY, busC,
                                       scalingX, scalingC)
            self.simulatorEndSignal.emit()
        except ValueError as exVE:
            warningMessage = '%s' % exVE
        except ZeroDivisionError as exZDE:
            warningMessage = '%s' % exZDE
        except RuntimeWarning as exRW:
            warningMessage = '%s' % exRW
        
        if warningMessage != '':
            QMessageBox.warning(self, 'Error on Simulation', 
                                "%s" % warningMessage, QMessageBox.Ok)
        else:        
            self.on_simulator_draw()
            # Clear the Figure
            
               
    
    def create_simulator_layout(self):
        # Create the mpl Figure and FigCanvas objects. 
        # 5x4 inches, 100 dots-per-inch
        #
        #self.dpi = 100
        #self.figSimulator = Figure((5.0, 4.0), dpi=self.dpi)
        self.figSimulator = Figure()
        self.canvasSimulator = FigureCanvas(self.figSimulator)
        self.canvasSimulator.setParent(self.main_frame)
        
        # Since we have only one plot, we can use add_axes 
        # instead of add_subplot, but then the subplot
        # configuration tool in the navigation toolbar wouldn't
        # work.
        #
        #self.axesSimulator = self.figSimulator.add_subplot(111)
        
        # Bind the 'pick' event for clicking on one of the bars
        #
        self.canvasSimulator.mpl_connect('pick_event', self.on_pick)
        
        # Create the navigation toolbar, tied to the canvas
        #
        self.mpl_toolbar_simulator = NavigationToolbar(self.canvasSimulator, self.main_frame)
        
        # Other GUI controls
        # 
        
        labelSimulatorEngine = QLabel('Simulation Engine:')
        self.comboSimulatorEngine = QComboBox()
        self.comboSimulatorEngine.addItem("Icarus Verilog")
        self.comboSimulatorEngine.addItem("GHDL")
        self.comboSimulatorEngine.setEditable(True) 
        self.comboSimulatorEngine.lineEdit().setReadOnly(True)
        self.comboSimulatorEngine.lineEdit().setAlignment(Qt.AlignRight)
        for ii in range(self.comboSimulatorEngine.count()):
            self.comboSimulatorEngine.setItemData(ii, Qt.AlignRight, Qt.TextAlignmentRole)
        self.connect(self.comboSimulatorEngine, SIGNAL('currentIndexChanged(int)'), self.on_simulator_modified)
        
        labelSimulatorLanguage = QLabel('HDL Language:')
        self.comboSimulatorLanguage = QComboBox()
        self.comboSimulatorLanguage.addItem("VHDL")
        self.comboSimulatorLanguage.addItem("Verilog")
        self.comboSimulatorLanguage.setEditable(True) 
        self.comboSimulatorLanguage.lineEdit().setReadOnly(True)
        self.comboSimulatorLanguage.lineEdit().setAlignment(Qt.AlignRight)
        for ii in range(self.comboSimulatorLanguage.count()):
            self.comboSimulatorLanguage.setItemData(ii, Qt.AlignRight, Qt.TextAlignmentRole)
        self.connect(self.comboSimulatorLanguage, SIGNAL('currentIndexChanged(int)'), self.on_language_modified)
        
        labelSimulatorScope = QLabel('Output Scope:')
        self.comboSimulatorScope = QComboBox()
        self.comboSimulatorScope.addItem("Time Plot")
        self.comboSimulatorScope.addItem("Error Plot")
        self.comboSimulatorScope.addItem("Power Spectrum")
        self.comboSimulatorScope.setEditable(True) 
        self.comboSimulatorScope.lineEdit().setReadOnly(True)
        self.comboSimulatorScope.lineEdit().setAlignment(Qt.AlignRight)
        for ii in range(self.comboSimulatorScope.count()):
            self.comboSimulatorScope.setItemData(ii, Qt.AlignRight, Qt.TextAlignmentRole)
        self.connect(self.comboSimulatorScope, SIGNAL('currentIndexChanged(int)'), self.on_simulator_draw)
             
        self.cbGridSimulator = QCheckBox("Show &Grid")
        self.cbGridSimulator.setChecked(True)
        self.connect(self.cbGridSimulator, SIGNAL('stateChanged(int)'), self.on_simulator_draw)
        
        
        # filter name
        labelFilterName = QLabel('Filter Name:')
        self.textFilterName = QLineEdit()
        # validate only characters and numbers
        regExpFilterName = QRegExp('^[a-zA-Z0-9]+$')
        validatorFilterName = QRegExpValidator(regExpFilterName)
        self.textFilterName.setValidator(validatorFilterName)
        self.textFilterName.setText('myfilter')
        self.textFilterName.setAlignment(Qt.AlignRight)
        self.connect(self.textFilterName, SIGNAL('editingFinished()'), self.on_simulator_modified)

        
        # Work directory
        labelWorkFolder = QLabel('Work Folder:')
        self.textWorkFolder = QLineEdit()
        self.pbWorkFolder = QPushButton('...')
        self.connect(self.pbWorkFolder, SIGNAL('clicked()'), self.on_explore_work_folder)
        # TODO: how to validate a path??
        #regExpWorkFolder = QRegExp('^[a-zA-Z0-9]+$')
        #validatorFilterName = QRegExpValidator(regExpWorkFolder)
        #self.textFilterName.setValidator(validatorFilterName)
        self.textWorkFolder.setText('.')
        self.textWorkFolder.setAlignment(Qt.AlignRight)
        self.connect(self.textWorkFolder, SIGNAL('editingFinished()'), self.on_simulator_modified)
        
        

        #
        # Layout with box sizers
        # 
        hboxSimulatorEngine = QHBoxLayout()
        hboxSimulatorEngine.addWidget(labelSimulatorEngine)
        hboxSimulatorEngine.addWidget(self.comboSimulatorEngine)
        hboxSimulatorEngine.setStretch(0,1)
        hboxSimulatorEngine.setStretch(1,1)
        
        hboxSimulatorLanguage = QHBoxLayout()
        hboxSimulatorLanguage.addWidget(labelSimulatorLanguage)
        hboxSimulatorLanguage.addWidget(self.comboSimulatorLanguage)
        hboxSimulatorLanguage.setStretch(0,1)
        hboxSimulatorLanguage.setStretch(1,1)
        
        hboxSimulatorScope = QHBoxLayout()
        hboxSimulatorScope.addWidget(labelSimulatorScope)
        hboxSimulatorScope.addWidget(self.comboSimulatorScope)
        hboxSimulatorScope.setStretch(0,1)
        hboxSimulatorScope.setStretch(1,1)
        
        hboxFilterName = QHBoxLayout()
        hboxFilterName.addWidget(labelFilterName)
        hboxFilterName.addWidget(self.textFilterName)
        hboxFilterName.setStretch(0,1)
        hboxFilterName.setStretch(1,1)
        
        hboxWorkFolder = QHBoxLayout()
        hboxWorkFolder.addWidget(labelWorkFolder)
        hboxWorkFolder.addWidget(self.textWorkFolder)
        hboxWorkFolder.addWidget(self.pbWorkFolder)
        hboxWorkFolder.setStretch(0,5)
        hboxWorkFolder.setStretch(1,4)
        hboxWorkFolder.setStretch(2,1)
        
     
        
        vboxSimulatorControl = QVBoxLayout()
        vboxSimulatorControl.addLayout(hboxFilterName)
        vboxSimulatorControl.addLayout(hboxWorkFolder)
        vboxSimulatorControl.addLayout(hboxSimulatorEngine)
        vboxSimulatorControl.addLayout(hboxSimulatorLanguage)
        
        groupBoxSimulatorControl = QGroupBox('Simulator Control')
        groupBoxSimulatorControl.setLayout(vboxSimulatorControl)
        groupBoxSimulatorControl.isCheckable()
        
        
        vboxMain = QVBoxLayout()
        vboxMain.addWidget(self.canvasSimulator)
        vboxMain.addWidget(self.mpl_toolbar_simulator)
        vboxMain.addWidget(self.cbGridSimulator)
        vboxMain.addLayout(hboxSimulatorScope)
        vboxMain.addWidget(groupBoxSimulatorControl)
        
        #self.on_simulator_draw()
               
        return vboxMain

    def on_explore_work_folder(self):
        dialogWorkFolder = QFileDialog()
        #dialogWorkFolder.setFileMode(QFileDialog.Directory)
        #dialogWorkFolder.setOption(QFileDialog.ShowDirsOnly, True)
        self.textWorkFolder.setText(dialogWorkFolder.getExistingDirectory(self, 
            'Select Work Folder', './'))
    
    def on_simulator_modified(self):
        print('on_simulator_modified')
        self.simulatorUpdatedSignal.emit()


    def on_language_modified(self):
        # TODO: provisional, check the structure change and force to DF1 Trans
        activeLanguage = str(self.comboSimulatorLanguage.currentText())
        if activeLanguage == 'VHDL':
            self.on_simulator_modified()
        else:
            QMessageBox.information(self, 'Simulation Language', 
                                 'Sorry, but %s is not supported yet!'
                                 % activeLanguage, 
                                 QMessageBox.Ok)
            self.comboSimulatorLanguage.setCurrentIndex(0)
 
